@@ -56,6 +56,9 @@ struct Expr : public Node {
 
   // is this expression the boolean literal “false”?
   virtual bool is_literal_false() const;
+
+  // is this expression side-effect free?
+  virtual bool is_pure() const = 0;
 };
 
 struct Ternary : public Expr {
@@ -72,10 +75,9 @@ struct Ternary : public Expr {
   bool constant() const final;
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   void validate() const final;
   std::string to_string() const final;
+  bool is_pure() const final;
 
   /* Note we do not override is_lvalue. Unlike in C, ternary expressions are not
    * considered lvalues.
@@ -93,6 +95,7 @@ struct BinaryExpr : public Expr {
 
   BinaryExpr *clone() const override = 0;
   bool constant() const final;
+  bool is_pure() const final;
 };
 
 struct BooleanBinaryExpr : public BinaryExpr {
@@ -101,6 +104,7 @@ struct BooleanBinaryExpr : public BinaryExpr {
     const location &loc_);
   BooleanBinaryExpr() = delete;
 
+  Ptr<TypeExpr> type() const final;
   void validate() const final;
 };
 
@@ -111,36 +115,58 @@ struct Implication : public BooleanBinaryExpr {
   Implication *clone() const final;
   virtual ~Implication() = default;
 
-  Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
+// logical OR
 struct Or : public BooleanBinaryExpr {
 
   Or(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_);
   virtual ~Or() = default;
   Or *clone() const final;
 
-  Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
+// logical AND
 struct And : public BooleanBinaryExpr {
 
   And(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_);
   virtual ~And() = default;
   And *clone() const final;
 
+  mpz_class constant_fold() const final;
+  std::string to_string() const final;
+};
+
+// An 'x & y' expression where a decision has not yet been made as to whether
+// the '&' is a logical AND or a bitwise AND. These nodes can only occur in the
+// AST prior to symbol resolution.
+struct AmbiguousAmp : public BinaryExpr {
+
+  AmbiguousAmp(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_);
+  virtual ~AmbiguousAmp() = default;
+  AmbiguousAmp *clone() const final;
+
+  Ptr<TypeExpr> type() const final;
+  mpz_class constant_fold() const final;
+  std::string to_string() const final;
+};
+
+// An 'x | y' expression where a decision has not yet been made as to whether
+// the '|' is a logical OR or a bitwise OR. These nodes can only occur in the
+// AST prior to symbol resolution.
+struct AmbiguousPipe : public BinaryExpr {
+
+  AmbiguousPipe(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_,
+    const location &loc_);
+  virtual ~AmbiguousPipe() = default;
+  AmbiguousPipe *clone() const final;
+
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
@@ -152,7 +178,8 @@ struct UnaryExpr : public Expr {
   UnaryExpr *clone() const override = 0;
   virtual ~UnaryExpr() = default;
 
-  bool constant() const final;
+  bool constant() const override;
+  bool is_pure() const final;
 };
 
 struct Not : public UnaryExpr {
@@ -163,8 +190,6 @@ struct Not : public UnaryExpr {
 
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   void validate() const final;
   std::string to_string() const final;
 };
@@ -185,8 +210,6 @@ struct Lt : public ComparisonBinaryExpr {
 
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
@@ -198,8 +221,6 @@ struct Leq : public ComparisonBinaryExpr {
 
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
@@ -211,8 +232,6 @@ struct Gt : public ComparisonBinaryExpr {
 
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
@@ -224,8 +243,6 @@ struct Geq : public ComparisonBinaryExpr {
 
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
@@ -245,8 +262,6 @@ struct Eq : public EquatableBinaryExpr {
 
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
@@ -258,8 +273,6 @@ struct Neq : public EquatableBinaryExpr {
 
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
@@ -268,6 +281,7 @@ struct ArithmeticBinaryExpr : public BinaryExpr {
   ArithmeticBinaryExpr(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_,
     const location &loc_);
 
+  Ptr<TypeExpr> type() const final;
   void validate() const final;
 };
 
@@ -277,10 +291,7 @@ struct Add : public ArithmeticBinaryExpr {
   virtual ~Add() = default;
   Add *clone() const final;
 
-  Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
@@ -290,10 +301,7 @@ struct Sub : public ArithmeticBinaryExpr {
   virtual ~Sub() = default;
   Sub *clone() const final;
 
-  Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
@@ -305,8 +313,18 @@ struct Negative : public UnaryExpr {
 
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
+  void validate() const final;
+  std::string to_string() const final;
+};
+
+struct Bnot : public UnaryExpr {
+
+  Bnot(const Ptr<Expr> &rhs_, const location &loc_);
+  virtual ~Bnot() = default;
+  Bnot *clone() const final;
+
+  Ptr<TypeExpr> type() const final;
+  mpz_class constant_fold() const final;
   void validate() const final;
   std::string to_string() const final;
 };
@@ -317,10 +335,7 @@ struct Mul : public ArithmeticBinaryExpr {
   virtual ~Mul() = default;
   Mul *clone() const final;
 
-  Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
@@ -330,10 +345,7 @@ struct Div : public ArithmeticBinaryExpr {
   virtual ~Div() = default;
   Div *clone() const final;
 
-  Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
@@ -343,13 +355,63 @@ struct Mod : public ArithmeticBinaryExpr {
   virtual ~Mod() = default;
   Mod *clone() const final;
 
-  Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
+struct Lsh : public ArithmeticBinaryExpr {
+
+  Lsh(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_);
+  virtual ~Lsh() = default;
+  Lsh *clone() const final;
+
+  mpz_class constant_fold() const final;
+  std::string to_string() const final;
+};
+
+struct Rsh : public ArithmeticBinaryExpr {
+
+  Rsh(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_);
+  virtual ~Rsh() = default;
+  Rsh *clone() const final;
+
+  mpz_class constant_fold() const final;
+  std::string to_string() const final;
+};
+
+// bitwise AND
+struct Band : public ArithmeticBinaryExpr {
+
+  Band(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_);
+  virtual ~Band() = default;
+  Band *clone() const final;
+
+  mpz_class constant_fold() const final;
+  std::string to_string() const final;
+};
+
+// bitwise OR
+struct Bor : public ArithmeticBinaryExpr {
+
+  Bor(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_);
+  virtual ~Bor() = default;
+  Bor *clone() const final;
+
+  mpz_class constant_fold() const final;
+  std::string to_string() const final;
+};
+
+struct Xor : public ArithmeticBinaryExpr {
+
+  Xor(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_);
+  virtual ~Xor() = default;
+  Xor *clone() const final;
+
+  mpz_class constant_fold() const final;
+  std::string to_string() const final;
+};
+
+
 struct ExprID : public Expr {
 
   std::string id;
@@ -363,14 +425,13 @@ struct ExprID : public Expr {
   bool constant() const final;
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   void validate() const final;
   bool is_lvalue() const final;
   bool is_readonly() const final;
   std::string to_string() const final;
   bool is_literal_true() const final;
   bool is_literal_false() const final;
+  bool is_pure() const final;
 };
 
 struct Field : public Expr {
@@ -386,12 +447,11 @@ struct Field : public Expr {
   bool constant() const final;
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   void validate() const final;
   bool is_lvalue() const final;
   bool is_readonly() const final;
   std::string to_string() const final;
+  bool is_pure() const final;
 };
 
 struct Element : public Expr {
@@ -407,12 +467,11 @@ struct Element : public Expr {
   bool constant() const final;
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   void validate() const final;
   bool is_lvalue() const final;
   bool is_readonly() const final;
   std::string to_string() const final;
+  bool is_pure() const final;
 };
 
 struct FunctionCall : public Expr {
@@ -432,10 +491,9 @@ struct FunctionCall : public Expr {
   bool constant() const final;
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   void validate() const final;
   std::string to_string() const final;
+  bool is_pure() const final;
 };
 
 struct Quantifier : public Node {
@@ -460,8 +518,6 @@ struct Quantifier : public Node {
     const location &loc_);
   virtual ~Quantifier() = default;
   Quantifier *clone() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   void validate() const final;
   std::string to_string() const;
 
@@ -475,6 +531,9 @@ struct Quantifier : public Node {
 
   // get the lower bound of this quantified expression as a C expression
   std::string lower_bound() const;
+
+  // is this side-effect free?
+  bool is_pure() const;
 };
 
 struct Exists : public Expr {
@@ -490,10 +549,9 @@ struct Exists : public Expr {
   bool constant() const final;
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   void validate() const final;
   std::string to_string() const final;
+  bool is_pure() const final;
 };
 
 struct Forall : public Expr {
@@ -509,15 +567,12 @@ struct Forall : public Expr {
   bool constant() const final;
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   void validate() const final;
   std::string to_string() const final;
+  bool is_pure() const final;
 };
 
-struct IsUndefined : public Expr {
-
-  Ptr<Expr> expr;
+struct IsUndefined : public UnaryExpr {
 
   IsUndefined(const Ptr<Expr> &expr_, const location &loc_);
   virtual ~IsUndefined() = default;
@@ -526,8 +581,6 @@ struct IsUndefined : public Expr {
   bool constant() const final;
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   void validate() const final;
   std::string to_string() const final;
 };
