@@ -29,9 +29,8 @@ struct Number : public Expr {
   bool constant() const final;
   Ptr<TypeExpr> type() const final;
   mpz_class constant_fold() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
+  bool is_pure() const final;
 };
 
 }
