@@ -26,9 +26,6 @@ struct Property : public Node {
     const location &loc_);
   Property *clone() const final;
   virtual ~Property() = default;
-
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
 };
 
 }
