@@ -53,9 +53,6 @@ struct TypeExpr : public Node {
   // Is this the type Boolean? Note that this only returns true for the actual
   // type Boolean, and not for TypeExprIDs that point at Boolean.
   virtual bool is_boolean() const;
-
-  __attribute__((deprecated("equatable_with() has been replaced by coerces_to()")))
-  bool equatable_with(const TypeExpr &other) const;
 };
 
 struct Range : public TypeExpr {
@@ -68,8 +65,6 @@ struct Range : public TypeExpr {
   virtual ~Range() = default;
 
   mpz_class count() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   bool is_simple() const final;
   void validate() const final;
 
@@ -88,8 +83,6 @@ struct Scalarset : public TypeExpr {
   virtual ~Scalarset() = default;
 
   mpz_class count() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   bool is_simple() const final;
   void validate() const final;
 
@@ -114,8 +107,6 @@ struct Enum : public TypeExpr {
   virtual ~Enum() = default;
 
   mpz_class count() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   bool is_simple() const final;
   void validate() const final;
 
@@ -136,8 +127,6 @@ struct Record : public TypeExpr {
 
   mpz_class width() const final;
   mpz_class count() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   std::string to_string() const final;
 };
 
@@ -153,8 +142,6 @@ struct Array : public TypeExpr {
 
   mpz_class width() const final;
   mpz_class count() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   void validate() const final;
   std::string to_string() const final;
 };
@@ -171,8 +158,6 @@ struct TypeExprID : public TypeExpr {
 
   mpz_class width() const final;
   mpz_class count() const final;
-  // __attribute__((deprecated("operator== will be removed in a future release")))
-  bool operator==(const Node &other) const final;
   bool is_simple() const final;
   Ptr<TypeExpr> resolve() const final;
   void validate() const final;
