@@ -1,4 +1,5 @@
 #include <cassert>
+#include <climits>
 #include <cstddef>
 #include <cstdint>
 #include <gmpxx.h>
@@ -54,11 +55,6 @@ bool Ternary::constant() const {
   return cond->constant() && lhs->constant() && rhs->constant();
 }
 
-bool Ternary::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Ternary*>(&other);
-  return o != nullptr && *cond == *o->cond && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 Ptr<TypeExpr> Ternary::type() const {
   // TODO: assert lhs and rhs are compatible types.
   return lhs->type();
@@ -78,6 +74,10 @@ std::string Ternary::to_string() const {
     + rhs->to_string() + ")";
 }
 
+bool Ternary::is_pure() const {
+  return cond->is_pure() && lhs->is_pure() && rhs->is_pure();
+}
+
 BinaryExpr::BinaryExpr(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_,
   const location &loc_):
   Expr(loc_), lhs(lhs_), rhs(rhs_) { }
@@ -86,9 +86,17 @@ bool BinaryExpr::constant() const {
   return lhs->constant() && rhs->constant();
 }
 
+bool BinaryExpr::is_pure() const {
+  return lhs->is_pure() && rhs->is_pure();
+}
+
 BooleanBinaryExpr::BooleanBinaryExpr(const Ptr<Expr> &lhs_,
   const Ptr<Expr> &rhs_, const location &loc_): BinaryExpr(lhs_, rhs_, loc_) { }
 
+Ptr<TypeExpr> BooleanBinaryExpr::type() const {
+  return Boolean;
+}
+
 void BooleanBinaryExpr::validate() const {
   if (!lhs->is_boolean())
     throw Error("left hand side of expression is not a boolean", lhs->loc);
@@ -105,19 +113,10 @@ Implication *Implication::clone() const {
   return new Implication(*this);
 }
 
-Ptr<TypeExpr> Implication::type() const {
-  return Boolean;
-}
-
 mpz_class Implication::constant_fold() const {
   return lhs->constant_fold() == 0 || rhs->constant_fold() != 0;
 }
 
-bool Implication::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Implication*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string Implication::to_string() const {
   return "(" + lhs->to_string() + " -> " + rhs->to_string() + ")";
 }
@@ -129,19 +128,10 @@ Or *Or::clone() const {
   return new Or(*this);
 }
 
-Ptr<TypeExpr> Or::type() const {
-  return Boolean;
-}
-
 mpz_class Or::constant_fold() const {
   return lhs->constant_fold() != 0 || rhs->constant_fold() != 0;
 }
 
-bool Or::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Or*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string Or::to_string() const {
   return "(" + lhs->to_string() + " | " + rhs->to_string() + ")";
 }
@@ -153,23 +143,60 @@ And *And::clone() const {
   return new And(*this);
 }
 
-Ptr<TypeExpr> And::type() const {
-  return Boolean;
-}
-
 mpz_class And::constant_fold() const {
   return lhs->constant_fold() != 0 && rhs->constant_fold() != 0;
 }
 
-bool And::operator==(const Node &other) const {
-  auto o = dynamic_cast<const And*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string And::to_string() const {
   return "(" + lhs->to_string() + " & " + rhs->to_string() + ")";
 }
 
+AmbiguousAmp::AmbiguousAmp(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_,
+  const location &loc_): BinaryExpr(lhs_, rhs_, loc_) { }
+
+AmbiguousAmp *AmbiguousAmp::clone() const {
+  return new AmbiguousAmp(*this);
+}
+
+Ptr<TypeExpr> AmbiguousAmp::type() const {
+  // we cannot retrieve the type of this expression because it has no certain
+  // type until we decide whether it is a logical AND or bitwise AND
+  throw Error("cannot retrieve the type of an unresolved '&' expression", loc);
+}
+
+mpz_class AmbiguousAmp::constant_fold() const {
+  // we cannot constant fold this if we do not yet know whether it is a logical
+  // AND or a bitwise AND
+  throw Error("cannot constant fold an unresolved '&' expression", loc);
+}
+
+std::string AmbiguousAmp::to_string() const {
+  return "(" + lhs->to_string() + " & " + rhs->to_string() + ")";
+}
+
+AmbiguousPipe::AmbiguousPipe(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_,
+  const location &loc_): BinaryExpr(lhs_, rhs_, loc_) { }
+
+AmbiguousPipe *AmbiguousPipe::clone() const {
+  return new AmbiguousPipe(*this);
+}
+
+Ptr<TypeExpr> AmbiguousPipe::type() const {
+  // we cannot retrieve the type of this expression because it has no certain
+  // type until we decide whether it is a logical OR or bitwise OR
+  throw Error("cannot retrieve the type of an unresolved '|' expression", loc);
+}
+
+mpz_class AmbiguousPipe::constant_fold() const {
+  // we cannot constant fold this if we do not yet know whether it is a logical
+  // OR or a bitwise OR
+  throw Error("cannot constant fold an unresolved '|' expression", loc);
+}
+
+std::string AmbiguousPipe::to_string() const {
+  return "(" + lhs->to_string() + " | " + rhs->to_string() + ")";
+}
+
 UnaryExpr::UnaryExpr(const Ptr<Expr> &rhs_, const location &loc_):
   Expr(loc_), rhs(rhs_) {
 }
@@ -178,6 +205,10 @@ bool UnaryExpr::constant() const {
   return rhs->constant();
 }
 
+bool UnaryExpr::is_pure() const {
+  return rhs->is_pure();
+}
+
 Not::Not(const Ptr<Expr> &rhs_, const location &loc_): UnaryExpr(rhs_, loc_) { }
 
 Not *Not::clone() const {
@@ -192,11 +223,6 @@ mpz_class Not::constant_fold() const {
   return rhs->constant_fold() == 0;
 }
 
-bool Not::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Not*>(&other);
-  return o != nullptr && *rhs == *o->rhs;
-}
-
 void Not::validate() const {
   if (!rhs->is_boolean())
     throw Error("argument to ! is not a boolean", rhs->loc);
@@ -242,11 +268,6 @@ mpz_class Lt::constant_fold() const {
   return lhs->constant_fold() < rhs->constant_fold();
 }
 
-bool Lt::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Lt*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string Lt::to_string() const {
   return "(" + lhs->to_string() + " < " + rhs->to_string() + ")";
 }
@@ -266,11 +287,6 @@ mpz_class Leq::constant_fold() const {
   return lhs->constant_fold() <= rhs->constant_fold();
 }
 
-bool Leq::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Leq*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string Leq::to_string() const {
   return "(" + lhs->to_string() + " <= " + rhs->to_string() + ")";
 }
@@ -290,11 +306,6 @@ mpz_class Gt::constant_fold() const {
   return lhs->constant_fold() > rhs->constant_fold();
 }
 
-bool Gt::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Gt*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string Gt::to_string() const {
   return "(" + lhs->to_string() + " > " + rhs->to_string() + ")";
 }
@@ -314,11 +325,6 @@ mpz_class Geq::constant_fold() const {
   return lhs->constant_fold() >= rhs->constant_fold();
 }
 
-bool Geq::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Geq*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string Geq::to_string() const {
   return "(" + lhs->to_string() + " >= " + rhs->to_string() + ")";
 }
@@ -346,11 +352,6 @@ mpz_class Eq::constant_fold() const {
   return lhs->constant_fold() == rhs->constant_fold();
 }
 
-bool Eq::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Eq*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string Eq::to_string() const {
   return "(" + lhs->to_string() + " = " + rhs->to_string() + ")";
 }
@@ -370,11 +371,6 @@ mpz_class Neq::constant_fold() const {
   return lhs->constant_fold() != rhs->constant_fold();
 }
 
-bool Neq::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Neq*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string Neq::to_string() const {
   return "(" + lhs->to_string() + " != " + rhs->to_string() + ")";
 }
@@ -399,6 +395,10 @@ void ArithmeticBinaryExpr::validate() const {
       loc);
 }
 
+Ptr<TypeExpr> ArithmeticBinaryExpr::type() const {
+  return Ptr<Range>::make(nullptr, nullptr, location());
+}
+
 Add::Add(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_):
   ArithmeticBinaryExpr(lhs_, rhs_, loc_) { }
 
@@ -406,19 +406,10 @@ Add *Add::clone() const {
   return new Add(*this);
 }
 
-Ptr<TypeExpr> Add::type() const {
-  return Ptr<Range>::make(nullptr, nullptr, location());
-}
-
 mpz_class Add::constant_fold() const {
   return lhs->constant_fold() + rhs->constant_fold();
 }
 
-bool Add::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Add*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string Add::to_string() const {
   return "(" + lhs->to_string() + " + " + rhs->to_string() + ")";
 }
@@ -430,19 +421,10 @@ Sub *Sub::clone() const {
   return new Sub(*this);
 }
 
-Ptr<TypeExpr> Sub::type() const {
-  return Ptr<Range>::make(nullptr, nullptr, location());
-}
-
 mpz_class Sub::constant_fold() const {
   return lhs->constant_fold() - rhs->constant_fold();
 }
 
-bool Sub::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Sub*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string Sub::to_string() const {
   return "(" + lhs->to_string() + " - " + rhs->to_string() + ")";
 }
@@ -467,15 +449,34 @@ mpz_class Negative::constant_fold() const {
   return -rhs->constant_fold();
 }
 
-bool Negative::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Negative*>(&other);
-  return o != nullptr && *rhs == *o->rhs;
-}
-
 std::string Negative::to_string() const {
   return "(-" + rhs->to_string() + ")";
 }
 
+Bnot::Bnot(const Ptr<Expr> &rhs_, const location &loc_):
+  UnaryExpr(rhs_, loc_) { }
+
+void Bnot::validate() const {
+  if (!isa<Range>(rhs->type()->resolve()))
+    throw Error("expression cannot be bitwise NOTed", rhs->loc);
+}
+
+Bnot *Bnot::clone() const {
+  return new Bnot(*this);
+}
+
+Ptr<TypeExpr> Bnot::type() const {
+  return Ptr<Range>::make(nullptr, nullptr, location());
+}
+
+mpz_class Bnot::constant_fold() const {
+  return ~rhs->constant_fold();
+}
+
+std::string Bnot::to_string() const {
+  return "(~" + rhs->to_string() + ")";
+}
+
 Mul::Mul(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_):
   ArithmeticBinaryExpr(lhs_, rhs_, loc_) { }
 
@@ -483,19 +484,10 @@ Mul *Mul::clone() const {
   return new Mul(*this);
 }
 
-Ptr<TypeExpr> Mul::type() const {
-  return Ptr<Range>::make(nullptr, nullptr, location());
-}
-
 mpz_class Mul::constant_fold() const {
   return lhs->constant_fold() * rhs->constant_fold();
 }
 
-bool Mul::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Mul*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string Mul::to_string() const {
   return "(" + lhs->to_string() + " * " + rhs->to_string() + ")";
 }
@@ -507,10 +499,6 @@ Div *Div::clone() const {
   return new Div(*this);
 }
 
-Ptr<TypeExpr> Div::type() const {
-  return Ptr<Range>::make(nullptr, nullptr, location());
-}
-
 mpz_class Div::constant_fold() const {
   mpz_class a = lhs->constant_fold();
   mpz_class b = rhs->constant_fold();
@@ -519,11 +507,6 @@ mpz_class Div::constant_fold() const {
   return a / b;
 }
 
-bool Div::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Div*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string Div::to_string() const {
   return "(" + lhs->to_string() + " / " + rhs->to_string() + ")";
 }
@@ -535,10 +518,6 @@ Mod *Mod::clone() const {
   return new Mod(*this);
 }
 
-Ptr<TypeExpr> Mod::type() const {
-  return Ptr<Range>::make(nullptr, nullptr, location());
-}
-
 mpz_class Mod::constant_fold() const {
   mpz_class a = lhs->constant_fold();
   mpz_class b = rhs->constant_fold();
@@ -547,15 +526,145 @@ mpz_class Mod::constant_fold() const {
   return a % b;
 }
 
-bool Mod::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Mod*>(&other);
-  return o != nullptr && *lhs == *o->lhs && *rhs == *o->rhs;
-}
-
 std::string Mod::to_string() const {
   return "(" + lhs->to_string() + " % " + rhs->to_string() + ")";
 }
 
+Lsh::Lsh(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_):
+  ArithmeticBinaryExpr(lhs_, rhs_, loc_) { }
+
+Lsh *Lsh::clone() const {
+  return new Lsh(*this);
+}
+
+// right shift an mpz value
+static mpz_class rshift(mpz_class a, mpz_class b);
+
+// left shift an mpz value
+static mpz_class lshift(mpz_class a, mpz_class b) {
+
+  // is this actually a right shift?
+  if (b < 0)
+    return rshift(a, -b);
+
+  // if the shift is beyond what we can do in one shot, recurse
+  while (!b.fits_ulong_p()) {
+    a = lshift(a, mpz_class(ULONG_MAX));
+    b -= ULONG_MAX;
+  }
+
+  // extract the shift value into a bit count
+  mp_bitcnt_t l = static_cast<mp_bitcnt_t>(b.get_ui());
+
+  // do a left shift using the GMP C API
+  mpz_t rop;
+  mpz_init(rop);
+  mpz_mul_2exp(rop, a.get_mpz_t(), l);
+
+  return mpz_class(rop);
+}
+
+static mpz_class rshift(mpz_class a, mpz_class b) {
+
+  // is this actually a left shift?
+  if (b < 0)
+    return lshift(a, -b);
+
+  // if the shift is beyond what we can do in one shot, recurse
+  while (!b.fits_ulong_p()) {
+    a = rshift(a, mpz_class(ULONG_MAX));
+    b -= ULONG_MAX;
+  }
+
+  // extract the shift value into a bit count
+  mp_bitcnt_t r = static_cast<mp_bitcnt_t>(b.get_ui());
+
+  // do a right shift using the GMP C API
+  mpz_t rop;
+  mpz_init(rop);
+  mpz_fdiv_q_2exp(rop, a.get_mpz_t(), r);
+
+  return mpz_class(rop);
+}
+
+mpz_class Lsh::constant_fold() const {
+  mpz_class a = lhs->constant_fold();
+  mpz_class b = rhs->constant_fold();
+  return lshift(a, b);
+}
+
+std::string Lsh::to_string() const {
+  return "(" + lhs->to_string() + " << " + rhs->to_string() + ")";
+}
+
+Rsh::Rsh(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_):
+  ArithmeticBinaryExpr(lhs_, rhs_, loc_) { }
+
+Rsh *Rsh::clone() const {
+  return new Rsh(*this);
+}
+
+mpz_class Rsh::constant_fold() const {
+  mpz_class a = lhs->constant_fold();
+  mpz_class b = rhs->constant_fold();
+  return rshift(a, b);
+}
+
+std::string Rsh::to_string() const {
+  return "(" + lhs->to_string() + " >> " + rhs->to_string() + ")";
+}
+
+Band::Band(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_):
+  ArithmeticBinaryExpr(lhs_, rhs_, loc_) { }
+
+Band *Band::clone() const {
+  return new Band(*this);
+}
+
+mpz_class Band::constant_fold() const {
+  mpz_class a = lhs->constant_fold();
+  mpz_class b = rhs->constant_fold();
+  return a & b;
+}
+
+std::string Band::to_string() const {
+  return "(" + lhs->to_string() + " & " + rhs->to_string() + ")";
+}
+
+Bor::Bor(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_):
+  ArithmeticBinaryExpr(lhs_, rhs_, loc_) { }
+
+Bor *Bor::clone() const {
+  return new Bor(*this);
+}
+
+mpz_class Bor::constant_fold() const {
+  mpz_class a = lhs->constant_fold();
+  mpz_class b = rhs->constant_fold();
+  return a | b;
+}
+
+std::string Bor::to_string() const {
+  return "(" + lhs->to_string() + " | " + rhs->to_string() + ")";
+}
+
+Xor::Xor(const Ptr<Expr> &lhs_, const Ptr<Expr> &rhs_, const location &loc_):
+  ArithmeticBinaryExpr(lhs_, rhs_, loc_) { }
+
+Xor *Xor::clone() const {
+  return new Xor(*this);
+}
+
+mpz_class Xor::constant_fold() const {
+  mpz_class a = lhs->constant_fold();
+  mpz_class b = rhs->constant_fold();
+  return a ^ b;
+}
+
+std::string Xor::to_string() const {
+  return "(" + lhs->to_string() + " ^ " + rhs->to_string() + ")";
+}
+
 ExprID::ExprID(const std::string &id_, const Ptr<ExprDecl> &value_,
   const location &loc_):
   Expr(loc_), id(id_), value(value_) {
@@ -594,23 +703,6 @@ mpz_class ExprID::constant_fold() const {
   throw Error("symbol \"" + id + "\" is not a constant", loc);
 }
 
-bool ExprID::operator==(const Node &other) const {
-  auto o = dynamic_cast<const ExprID*>(&other);
-  if (o == nullptr)
-    return false;
-  if (id != o->id)
-    return false;
-  if (value == nullptr) {
-    if (o->value != nullptr)
-      return false;
-  } else if (o->value == nullptr) {
-    return false;
-  } else if (*value != *o->value) {
-    return false;
-  }
-  return true;
-}
-
 void ExprID::validate() const {
   if (value == nullptr)
     throw Error("unresolved expression \"" + id + "\"", loc);
@@ -642,6 +734,10 @@ bool ExprID::is_literal_false() const {
   return id == "false";
 }
 
+bool ExprID::is_pure() const {
+  return true;
+}
+
 Field::Field(const Ptr<Expr> &record_, const std::string &field_,
   const location &loc_):
   Expr(loc_), record(record_), field(field_) {
@@ -661,7 +757,12 @@ Ptr<TypeExpr> Field::type() const {
   const Ptr<TypeExpr> resolved = root->resolve();
 
   auto r = dynamic_cast<const Record*>(resolved.get());
-  assert(r != nullptr && "invalid left hand side of field expression");
+
+  // if we are called before types have been resolved (or during type
+  // resolution on a malformed model), it is possible that the root is not a
+  // record
+  if (r == nullptr)
+    throw Error("invalid left hand side of field expression", loc);
 
   for (const Ptr<VarDecl> &f : r->fields) {
     if (f->name == field)
@@ -675,11 +776,6 @@ mpz_class Field::constant_fold() const {
   throw Error("field expression used in constant", loc);
 }
 
-bool Field::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Field*>(&other);
-  return o != nullptr && *record == *o->record && field == o->field;
-}
-
 void Field::validate() const {
 
   const Ptr<TypeExpr> root = record->type()->resolve();
@@ -709,6 +805,10 @@ std::string Field::to_string() const {
   return record->to_string() + "." + field;
 }
 
+bool Field::is_pure() const {
+  return true;
+}
+
 Element::Element(const Ptr<Expr> &array_, const Ptr<Expr> &index_,
   const location &loc_):
   Expr(loc_), array(array_), index(index_) {
@@ -725,8 +825,12 @@ bool Element::constant() const {
 Ptr<TypeExpr> Element::type() const {
   const Ptr<TypeExpr> t = array->type()->resolve();
   const Array *a = dynamic_cast<const Array*>(t.get());
-  assert(a != nullptr &&
-    "array reference based on something that is not an array");
+
+  // if we are called during symbol resolution on a malformed expression, our
+  // left hand side may not be an array
+  if (a == nullptr)
+    throw Error("array reference based on something that is not an array", loc);
+
   return a->element_type;
 }
 
@@ -734,11 +838,6 @@ mpz_class Element::constant_fold() const {
   throw Error("array element used in constant", loc);
 }
 
-bool Element::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Element*>(&other);
-  return o != nullptr && *array == *o->array && *index == *o->index;
-}
-
 void Element::validate() const {
 
   const Ptr<TypeExpr> t = array->type()->resolve();;
@@ -764,6 +863,10 @@ std::string Element::to_string() const {
   return array->to_string() + "[" + index->to_string() + "]";
 }
 
+bool Element::is_pure() const {
+  return true;
+}
+
 FunctionCall::FunctionCall(const std::string &name_,
   const std::vector<Ptr<Expr>> &arguments_, const location &loc_):
   Expr(loc_), name(name_), arguments(arguments_) { }
@@ -795,27 +898,6 @@ mpz_class FunctionCall::constant_fold() const {
   throw Error("function call used in a constant", loc);
 }
 
-bool FunctionCall::operator==(const Node &other) const {
-  auto o = dynamic_cast<const FunctionCall*>(&other);
-  if (o == nullptr)
-    return false;
-  if (name != o->name)
-    return false;
-  if (function == nullptr) {
-    if (o->function != nullptr)
-      return false;
-  } else if (o->function == nullptr) {
-    return false;
-  } else if (*function != *o->function) {
-    return false;
-  }
-  if (!vector_eq(arguments, o->arguments))
-    return false;
-  if (within_procedure_call != o->within_procedure_call)
-    return false;
-  return true;
-}
-
 void FunctionCall::validate() const {
   if (function == nullptr)
     throw Error("unknown function call \"" + name + "\"", loc);
@@ -876,6 +958,26 @@ std::string FunctionCall::to_string() const {
   return s;
 }
 
+bool FunctionCall::is_pure() const {
+
+  // If this is a recursive call within a function, it may have no referent. In
+  // this case, conservatively assume it is impure.
+  if (function == nullptr)
+    return false;
+
+  // if the function itself has side effects, the function call is impure
+  if (!function->is_pure())
+    return false;
+
+  // if any of the parameters have side effects, the function call is impure
+  for (const Ptr<Expr> &a : arguments) {
+    if (!a->is_pure())
+      return false;
+  }
+
+  return true;
+}
+
 Quantifier::Quantifier(const std::string &name_, const Ptr<TypeExpr> &type_,
   const location &loc_):
   Node(loc_), name(name_), type(type_),
@@ -898,47 +1000,6 @@ Quantifier *Quantifier::clone() const {
   return new Quantifier(*this);
 }
 
-bool Quantifier::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Quantifier*>(&other);
-  if (o == nullptr)
-    return false;
-  if (name != o->name)
-    return false;
-  if (type == nullptr) {
-    if (o->type != nullptr)
-      return false;
-  } else if (o->type == nullptr) {
-    return false;
-  } else if (*type != *o->type) {
-    return false;
-  }
-  if (from == nullptr) {
-    if (o->from != nullptr)
-      return false;
-  } else if (o->from == nullptr) {
-    return false;
-  } else if (*from != *o->from) {
-    return false;
-  }
-  if (to == nullptr) {
-    if (o->to != nullptr)
-      return false;
-  } else if (o->to == nullptr) {
-    return false;
-  } else if (*to != *o->to) {
-    return false;
-  }
-  if (step == nullptr) {
-    if (o->step != nullptr)
-      return false;
-  } else if (o->step == nullptr) {
-    return false;
-  } else if (*step != *o->step) {
-    return false;
-  }
-  return true;
-}
-
 void Quantifier::validate() const {
 
   bool from_const = from != nullptr && from->constant();
@@ -1036,6 +1097,35 @@ std::string Quantifier::lower_bound() const {
   return "VALUE_C(" + from->constant_fold().get_str() + ")";
 }
 
+bool Quantifier::is_pure() const {
+
+  if (type != nullptr) {
+    const Ptr<TypeExpr> t = type->resolve();
+
+    if (auto r = dynamic_cast<const Range*>(t.get()))
+      return r->min->is_pure() && r->max->is_pure();
+
+    if (auto s = dynamic_cast<const Scalarset*>(t.get()))
+      return s->bound->is_pure();
+
+    assert(dynamic_cast<const Enum*>(t.get()) != nullptr &&
+      "complex type encountered in quantifier");
+
+    return true;
+  }
+
+  if (from != nullptr && !from->is_pure())
+    return false;
+
+  if (to != nullptr && !to->is_pure())
+    return false;
+
+  if (step != nullptr && !step->is_pure())
+    return false;
+
+  return true;
+}
+
 Exists::Exists(const Quantifier &quantifier_, const Ptr<Expr> &expr_,
   const location &loc_):
   Expr(loc_), quantifier(quantifier_), expr(expr_) { }
@@ -1052,11 +1142,6 @@ Ptr<TypeExpr> Exists::type() const {
   return Boolean;
 }
 
-bool Exists::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Exists*>(&other);
-  return o != nullptr && quantifier == o->quantifier && *expr == *o->expr;
-}
-
 mpz_class Exists::constant_fold() const {
   throw Error("exists expression used in constant", loc);
 }
@@ -1071,6 +1156,10 @@ std::string Exists::to_string() const {
     + " endexists";
 }
 
+bool Exists::is_pure() const {
+  return quantifier.is_pure() && expr->is_pure();
+}
+
 Forall::Forall(const Quantifier &quantifier_, const Ptr<Expr> &expr_,
   const location &loc_):
   Expr(loc_), quantifier(quantifier_), expr(expr_) { }
@@ -1091,11 +1180,6 @@ mpz_class Forall::constant_fold() const {
   throw Error("forall expression used in constant", loc);
 }
 
-bool Forall::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Forall*>(&other);
-  return o != nullptr && quantifier == o->quantifier && *expr == *o->expr;
-}
-
 void Forall::validate() const {
   if (!expr->is_boolean())
     throw Error("expression in forall is not boolean", expr->loc);
@@ -1106,8 +1190,12 @@ std::string Forall::to_string() const {
     + " endforall";
 }
 
+bool Forall::is_pure() const {
+  return quantifier.is_pure() && expr->is_pure();
+}
+
 IsUndefined::IsUndefined(const Ptr<Expr> &expr_, const location &loc_):
-  Expr(loc_), expr(expr_) { }
+  UnaryExpr(expr_, loc_) { }
 
 IsUndefined *IsUndefined::clone() const {
   return new IsUndefined(*this);
@@ -1125,29 +1213,20 @@ mpz_class IsUndefined::constant_fold() const {
   throw Error("isundefined used in constant", loc);
 }
 
-bool IsUndefined::operator==(const Node &other) const {
-  auto o = dynamic_cast<const IsUndefined*>(&other);
-  if (o == nullptr)
-    return false;
-  if (*expr != *o->expr)
-    return false;
-  return true;
-}
-
 void IsUndefined::validate() const {
 
-  if (!expr->is_lvalue())
+  if (!rhs->is_lvalue())
     throw Error("non-lvalue expression cannot be used in isundefined",
-      expr->loc);
+      rhs->loc);
 
-  const Ptr<TypeExpr> t = expr->type();
+  const Ptr<TypeExpr> t = rhs->type();
   if (!t->is_simple())
-    throw Error("complex type used in isundefined", expr->loc);
+    throw Error("complex type used in isundefined", rhs->loc);
 
 }
 
 std::string IsUndefined::to_string() const {
-  return "isundefined(" + expr->to_string() + ")";
+  return "isundefined(" + rhs->to_string() + ")";
 }
 
 }
