@@ -12,9 +12,4 @@ Property *Property::clone() const {
   return new Property(*this);
 }
 
-bool Property::operator==(const Node &other) const {
-  auto o = dynamic_cast<const Property*>(&other);
-  return o != nullptr && category == o->category && *expr == *o->expr;
-}
-
 }
