@@ -151,6 +151,18 @@ void XMLPrinter::visit_assignment(const Assignment &n) {
   o << "</assignment>";
 }
 
+void XMLPrinter::visit_band(const Band &n) {
+  visit_bexpr("band", n);
+}
+
+void XMLPrinter::visit_bnot(const Bnot &n) {
+  visit_uexpr("bnot", n);
+}
+
+void XMLPrinter::visit_bor(const Bor &n) {
+  visit_bexpr("bor", n);
+}
+
 void XMLPrinter::visit_clear(const Clear &n) {
   sync_to(n);
   o << "<clear ";
@@ -423,20 +435,17 @@ void XMLPrinter::visit_implication(const Implication &n) {
 }
 
 void XMLPrinter::visit_isundefined(const IsUndefined &n) {
-  sync_to(n);
-  o << "<isundefined ";
-  add_location(n);
-  o << ">";
-  sync_to(*n.expr);
-  dispatch(*n.expr);
-  sync_to(n.loc.end);
-  o << "</isundefined>";
+  visit_uexpr("isundefined", n);
 }
 
 void XMLPrinter::visit_leq(const Leq &n) {
   visit_bexpr("leq", n);
 }
 
+void XMLPrinter::visit_lsh(const Lsh &n) {
+  visit_bexpr("lsh", n);
+}
+
 void XMLPrinter::visit_lt(const Lt &n) {
   visit_bexpr("lt", n);
 }
@@ -663,6 +672,10 @@ void XMLPrinter::visit_return(const Return &n) {
   o << "</return>";
 }
 
+void XMLPrinter::visit_rsh(const Rsh &n) {
+  visit_bexpr("rsh", n);
+}
+
 void XMLPrinter::visit_ruleset(const Ruleset &n) {
   sync_to(n);
   o << "<ruleset name=\"" << xml_escape(n.name) << "\" ";
@@ -924,6 +937,10 @@ void XMLPrinter::visit_while(const While &n) {
   o << "</while>";
 }
 
+void XMLPrinter::visit_xor(const Xor &n) {
+  visit_bexpr("xor", n);
+}
+
 XMLPrinter::~XMLPrinter() {
   sync_to();
   o << "</unit>\n";
