@@ -13,6 +13,7 @@
 #include <iostream>
 #include "log.h"
 #include <memory>
+#include "optimise-field-ordering.h"
 #include "options.h"
 #include "resources.h"
 #include <rumur/rumur.h>
@@ -57,7 +58,9 @@ static void parse_args(int argc, char **argv) {
       OPT_MONOPOLISE,
       OPT_OUTPUT_FORMAT,
       OPT_PACK_STATE,
+      OPT_REORDER_FIELDS,
       OPT_SANDBOX,
+      OPT_SCALARSET_SCHEDULES,
       OPT_SMT_ARG,
       OPT_SMT_BITVECTORS,
       OPT_SMT_BUDGET,
@@ -86,7 +89,9 @@ static void parse_args(int argc, char **argv) {
       { "output-format", required_argument, 0, OPT_OUTPUT_FORMAT },
       { "pack-state", required_argument, 0, OPT_PACK_STATE },
       { "quiet", no_argument, 0, 'q' },
+      { "reorder-fields", required_argument, 0, OPT_REORDER_FIELDS },
       { "sandbox", required_argument, 0, OPT_SANDBOX },
+      { "scalarset-schedules", required_argument, 0, OPT_SCALARSET_SCHEDULES },
       { "set-capacity", required_argument, 0, 's' },
       { "set-expand-threshold", required_argument, 0, 'e' },
       { "smt-arg", required_argument, 0, OPT_SMT_ARG },
@@ -307,6 +312,18 @@ static void parse_args(int argc, char **argv) {
         }
         break;
 
+      case OPT_SCALARSET_SCHEDULES: // --scalarset-schedules ...
+        if (strcmp(optarg, "on") == 0) {
+          options.scalarset_schedules = true;
+        } else if (strcmp(optarg, "off") == 0) {
+          options.scalarset_schedules = false;
+        } else {
+          std::cerr << "invalid argument to --scalarset-schedules, \"" << optarg
+            << "\"\n";
+          exit(EXIT_FAILURE);
+        }
+        break;
+
       case OPT_MAX_ERRORS: { // --max-errors ...
         bool valid = true;
         try {
@@ -375,6 +392,18 @@ static void parse_args(int argc, char **argv) {
         options.value_type = optarg;
         break;
 
+      case OPT_REORDER_FIELDS: // --reorder-fields ...
+        if (strcmp(optarg, "on") == 0) {
+          options.reorder_fields = true;
+        } else if (strcmp(optarg, "off") == 0) {
+          options.reorder_fields = false;
+        } else {
+          std::cerr << "invalid argument to --reorder-fields, \"" << optarg
+            << "\"\n";
+          exit(EXIT_FAILURE);
+        }
+        break;
+
       case OPT_SMT_ARG: // --smt-arg ...
         options.smt.args.emplace_back(optarg);
         if (options.smt.simplification == SmtSimplification::AUTO) {
@@ -588,6 +617,7 @@ int main(int argc, char **argv) {
   parse_args(argc, argv);
 
   // Parse input model
+  *debug << "parsing input model...\n";
   Ptr<Model> m;
   try {
     m = parse(in == nullptr ? std::cin : *in);
@@ -604,11 +634,14 @@ int main(int argc, char **argv) {
   /* Re-index the model (assign unique identifiers to each node that are used in
    * generation of the verifier).
    */
+  *debug << "re-indexing...\n";
   m->reindex();
 
   // resolve symbolic references and validate the model
   try {
+    *debug << "resolving symbols...\n";
     resolve_symbols(*m);
+    *debug << "validating AST...\n";
     validate(*m);
   } catch (Error &e) {
     std::cerr << white() << bold() << input_filename << ":" << e.loc << ":"
@@ -624,6 +657,7 @@ int main(int argc, char **argv) {
 
   // run SMT simplification if the user enabled it
   if (options.smt.simplification == SmtSimplification::ON) {
+    *debug << "SMT simplification...\n";
     try {
       smt::simplify(*m);
     } catch (smt::BudgetExhausted&) {
@@ -635,7 +669,14 @@ int main(int argc, char **argv) {
     }
   }
 
+  // re-order fields to optimise access to them
+  if (options.reorder_fields) {
+    *debug << "optimising field ordering...\n";
+    optimise_field_ordering(*m);
+  }
+
   // get value_t to use in the checker
+  *debug << "determining value_t type...\n";
   std::pair<ValueType, ValueType> value_types;
   try {
     value_types = get_value_type(options.value_type, *m);
@@ -645,6 +686,7 @@ int main(int argc, char **argv) {
     return EXIT_FAILURE;
   }
 
+  *debug << "generating verifier...\n";
   assert(out != nullptr);
   if (output_checker(*out, *m, value_types) != 0)
     return EXIT_FAILURE;
