@@ -1,6 +1,7 @@
 #pragma once
 
 #include <cstddef>
+#include <gmpxx.h>
 #include <rumur/rumur.h>
 #include <string>
 
@@ -9,11 +10,11 @@ bool isa(const U ptr) {
   return ptr != nullptr && dynamic_cast<const T*>(&*ptr) != nullptr;
 }
 
-// escape a string for the purposes of outputting it to a C source file
-std::string escape(const std::string &s);
-
 // get a C source code string for this expression
 std::string to_C_string(const rumur::Expr &expr);
 
 // get a C source code string for this location
 std::string to_C_string(const rumur::location &location);
+
+// how many bits are required to store `v` unique values?
+mpz_class bit_width(const mpz_class &v);
