@@ -13,7 +13,7 @@ import shutil
 import subprocess as sp
 import sys
 import tempfile
-from typing import Optional, Tuple
+from typing import Any, Dict, Optional, Tuple
 
 CPUS = multiprocessing.cpu_count()
 
@@ -51,116 +51,18 @@ def pr(s: str) -> None:
   sys.stdout.flush()
   print_lock.release()
 
-def which(cmd: str) -> Optional[str]:
-  try:
-    return sp.check_output(['which', cmd], stderr=sp.DEVNULL,
-      universal_newlines=True).strip()
-  except sp.CalledProcessError:
-    return None
-
-def run(args: [str], stdin: Optional[str] = None) -> Tuple[int, str, str]:
+def run(args: [str], env: Dict[str, str], stdin: Optional[str] = None
+    ) -> Tuple[int, str, str]:
   if stdin is not None:
     stdin = enc(stdin)
-  p = sp.run(args, stdout=sp.PIPE, stderr=sp.PIPE, input=stdin)
+  p = sp.run(args, stdout=sp.PIPE, stderr=sp.PIPE, input=stdin, env=env)
   return p.returncode, dec(p.stdout), dec(p.stderr)
 
-# C compiler
-CC = os.environ.get('CC', which('cc'))
-
-def has_mcx16() -> bool:
-  'does the compiler support -mcx16?'
-
-  code = 'int main(void) {\n' \
-         '  return 0;\n' \
-         '}\n'
-
-  args = [CC, '-x', 'c', '-std=c11', '-mcx16', '-', '-o', os.devnull]
-  ret, _, _ = run(args, code)
-  return ret == 0
-
-HAS_MCX16 = has_mcx16()
-
-# initial flags to pass to our C compiler
-C_FLAGS = ['-x', 'c', '-std=c11', '-Werror=format', '-Werror=sign-compare',
-  '-Werror=type-limits'] + (['-mcx16'] if HAS_MCX16 else [])
-
-def needs_libatomic() -> bool:
-  'does the toolchain need -latomic to support dword CAS?'
-
-  code = '#include <stdint.h>\n' \
-         '\n' \
-         'int main(void) {\n' \
-         '#if __SIZEOF_POINTER__ <= 4\n' \
-         '  uint64_t target = 0;\n' \
-         '#elif __SIZEOF_POINTER__ <= 8\n' \
-         '  unsigned __int128 target = 0;\n' \
-         '#endif\n' \
-         '  return (int)__sync_val_compare_and_swap(&target, 0, 1);\n' \
-         '}\n'
-
-  args = [CC, '-x', 'c', '-std=c11', '-', '-o', os.devnull]
-  if HAS_MCX16:
-    args.append('-mcx16')
-  ret, _, _ = run(args, code)
-  return ret != 0
-
-NEEDS_LIBATOMIC = needs_libatomic()
-
 VERIFIER_RNG = os.path.abspath(os.path.join(os.path.dirname(__file__),
   '../misc/verifier.rng'))
 MURPHI2XML_RNG = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'misc',
       'murphi2xml.rng'))
 
-SECCOMP_SUPPORTED = os.path.abspath(os.path.join(os.path.dirname(__file__),
-  '../misc/seccomp-supported.sh'))
-
-def has_sandbox() -> bool:
-  'whether the current platform has sandboxing support for the verifier'
-
-  if platform.system() == 'Darwin':
-    return True
-
-  if platform.system() == 'FreeBSD':
-    return True
-
-  if platform.system() == 'Linux':
-    ret, _, _ = run([SECCOMP_SUPPORTED])
-    return ret == 0
-
-  if platform.system() == 'OpenBSD':
-    return True
-
-  return False
-
-HAS_SANDBOX = has_sandbox()
-
-def smt_args(bv: bool = False) -> [str]:
-  'get SMT arguments for an available solver'
-
-  # preference 1: Z3
-  if which('z3') is not None:
-    # we leave a blank logic here, as Z3 performs best when not given a logic
-    args = ['--smt-path', 'z3', '--smt-arg=-smt2', '--smt-arg=-in']
-    if bv:
-      args += ['--smt-bitvectors', 'on']
-    return args
-
-  # preference 2: CVC4
-  if which('cvc4') is not None:
-    args = ['--smt-path', 'cvc4', '--smt-arg=--lang=smt2',
-      '--smt-arg=--rewrite-divk']
-    if bv:
-      args += ['--smt-prelude', '(set-logic AUFBV)', '--smt-bitvectors', 'on']
-    else:
-      args += ['--smt-prelude', '(set-logic AUFLIA)']
-    return args
-
-  # otherwise, give up
-  return []
-
-SMT_ARGS = smt_args()
-SMT_BV_ARGS = smt_args(True)
-
 class TemporaryDirectory(object):
   'an mkdtemp() that cleans up after itself'
   def __init__(self):
@@ -172,8 +74,6 @@ class TemporaryDirectory(object):
     if self.tmp is not None:
       shutil.rmtree(self.tmp)
 
-HAS_VALGRIND = which('valgrind') is not None
-
 class Result(abc.ABC): pass
 class Skip(Result):
   def __init__(self, reason: str): self.reason = reason
@@ -181,6 +81,13 @@ class Fail(Result):
   def __init__(self, output: str): self.output = output
 
 class Test(abc.ABC):
+  def __init__(self, config: Dict[str, Any]):
+    self.config = config # save config in case child wants it
+  def env(self) -> Dict[str, str]:
+    'construct an environment for running external programs'
+    env = {k: v for k, v in os.environ.items()}
+    env.update({k: str(v) for k, v in self.config.items()})
+    return env
   @abc.abstractmethod
   def description(self) -> str: raise NotImplementedError
   @abc.abstractmethod
@@ -191,7 +98,8 @@ TWEAK_LINE = re.compile(r'\s*--\s*(?P<key>[a-zA-Z_]\w*)\s*:(?P<value>.*)$')
 
 class Tweakable(Test):
   'a test case that can take extra customisation via comment lines'
-  def __init__(self):
+  def __init__(self, config: Dict[str, Any]):
+    super().__init__(config)
     # default options
     self.rumur_flags = []
     self.rumur_exit_code = 0
@@ -211,20 +119,19 @@ class Tweakable(Test):
 
 class Model(Tweakable):
   def __init__(self, model: str, debug: bool, optimised: bool, \
-      multithreaded: bool, xml: bool):
-    super().__init__()
+      multithreaded: bool, xml: bool, config: Dict[str, Any]):
+    super().__init__(config)
     self.model = model
     self.debug = debug
     self.optimised = optimised
     self.multithreaded = multithreaded
     self.xml = xml
-    self.valgrind = HAS_VALGRIND
   def description(self) -> str:
-    return f'{"D" if self.debug         else " "}' \
-           f'{"O" if self.optimised     else " "}' \
-           f'{"M" if self.multithreaded else " "}' \
-           f'{"X" if self.xml           else " "}' \
-           f'{"V" if HAS_VALGRIND       else " "}' \
+    return f'{"D" if self.debug                  else " "}' \
+           f'{"O" if self.optimised              else " "}' \
+           f'{"M" if self.multithreaded          else " "}' \
+           f'{"X" if self.xml                    else " "}' \
+           f'{"V" if self.config["HAS_VALGRIND"] else " "}' \
            f' {os.path.basename(self.model)}'
   def run(self) -> Result:
 
@@ -240,13 +147,13 @@ class Model(Tweakable):
       + (['--threads', '1'] if not self.multithreaded else [])           \
       + self.rumur_flags
 
-    if HAS_VALGRIND:
+    if self.config['HAS_VALGRIND']:
       args = ['valgrind', '--leak-check=full', '--show-leak-kinds=all',
         '--error-exitcode=42'] + args
 
     # call rumur
-    ret, stdout, stderr = run(args)
-    if HAS_VALGRIND:
+    ret, stdout, stderr = run(args, self.env())
+    if self.config['HAS_VALGRIND']:
       if ret == 42:
         return Fail(f'Memory leak:\n{stdout}{stderr}')
     if ret != self.rumur_exit_code:
@@ -261,18 +168,19 @@ class Model(Tweakable):
 
       # build up arguments to call the C compiler
       model_bin = os.path.join(tmp, 'model.exe')
-      args = [CC] + C_FLAGS + ['-o', model_bin, '-', '-lpthread']
+      args = [self.config['CC']] + self.config['C_FLAGS'] + ['-o', model_bin,
+        '-', '-lpthread']
 
-      if NEEDS_LIBATOMIC:
+      if self.config['NEEDS_LIBATOMIC']:
         args.append('-latomic')
 
       # call the C compiler
-      ret, stdout, stderr = run(args, model_c)
+      ret, stdout, stderr = run(args, self.env(), model_c)
       if ret != 0:
         return Fail(f'C compilation failed:\n{stdout}{stderr}')
 
       # now run the model itself
-      ret, stdout, stderr = run([model_bin])
+      ret, stdout, stderr = run([model_bin], self.env())
       if ret != self.checker_exit_code:
         return Fail(f'Unexpected checker exit status {ret}:\n{stdout}{stderr}')
 
@@ -291,44 +199,146 @@ class Model(Tweakable):
 
       model_xml = stdout
 
-      if which('xmllint') is None: return Skip('xmllint not available')
+      if not self.config['HAS_XMLLINT']: return Skip('xmllint not available')
 
       # validate the XML
       args = ['xmllint', '--relaxng', VERIFIER_RNG, '--noout', '-']
-      ret, stdout, stderr = run(args, model_xml)
+      ret, stdout, stderr = run(args, self.env(), model_xml)
       if ret != 0:
         return Fail( 'Failed to XML-validate machine reachable output:\n'
                     f'{stdout}{stderr}')
 
 class Executable(Test):
-  def __init__(self, exe: str):
+  def __init__(self, exe: str, config: Dict[str, Any]):
+    super().__init__(config)
     self.exe = exe
   def description(self) -> str: return f'----- exec {os.path.basename(self.exe)}'
   def run(self) -> Result:
-    ret, stdout, stderr = run(self.exe)
+    ret, stdout, stderr = run(self.exe, self.env())
     output = f'{stdout}{stderr}'
     return None                 if ret == 0 else \
            Skip(output.strip()) if ret == 125 else \
            Fail(output)
 
-class Murphi2XMLTest(Tweakable):
-  def __init__(self, model: str):
-    super().__init__()
+class Murphi2CTest(Tweakable):
+  def __init__(self, model: str, config: Dict[str, Any]):
+    super().__init__(config)
     self.model = model
     self.xml = False # dummy setting that tests might reference
   def description(self) -> str:
-    return f'----{"V" if HAS_VALGRIND else " "} ' \
+    return f'----{"V" if self.config["HAS_VALGRIND"] else " "} ' \
+           f'murphi2c {os.path.basename(self.model)}'
+  def run(self) -> Result:
+
+    self.apply_options(self.model)
+
+    # there is no C equivalent of isundefined, because an implicit assumption in
+    # the C representation is that you do not rely on undefined values
+    with open(self.model, 'rt', encoding='utf-8') as f:
+      should_fail = re.search(r'\bisundefined\b', f.read()) is not None
+
+    args = ['murphi2c', self.model]
+    if self.config['HAS_VALGRIND']:
+      args = ['valgrind', '--leak-check=full', '--show-leak-kinds=all',
+        '--error-exitcode=42'] + args
+    ret, stdout, stderr = run(args, self.env())
+    if self.config['HAS_VALGRIND']:
+      if ret == 42:
+        return Fail(f'Memory leak:\n{stdout}{stderr}')
+
+    # if rumur was expected to reject this model, we allow murphi2c to fail
+    if self.rumur_exit_code == 0 and not should_fail and ret != 0:
+      return Fail(f'Unexpected murphi2c exit status {ret}:\n{stdout}{stderr}')
+
+    if should_fail and ret == 0:
+      return Fail(f'Unexpected murphi2c exit status {ret}:\n{stdout}{stderr}')
+
+    if ret != 0:
+      return None
+
+    # ask the C compiler if this is valid
+    args = [self.config['CC']] + self.config['C_FLAGS'] + ['-c', '-o',
+      os.devnull, '-']
+    ret, stdout, stderr = run(args, self.env(), stdout)
+    if ret != 0:
+      return Fail(f'C compilation failed:\n{stdout}{stderr}')
+
+class Murphi2CHeaderTest(Tweakable):
+  def __init__(self, model: str, config: Dict[str, Any]):
+    super().__init__(config)
+    self.model = model
+    self.xml = False # dummy setting that tests might reference
+  def description(self) -> str:
+    return f'----{"V" if self.config["HAS_VALGRIND"] else " "} ' \
+           f'murphi2c --header {os.path.basename(self.model)}'
+  def run(self) -> Result:
+
+    self.apply_options(self.model)
+
+    # there is no C equivalent of isundefined, because an implicit assumption in
+    # the C representation is that you do not rely on undefined values
+    with open(self.model, 'rt', encoding='utf-8') as f:
+      should_fail = re.search(r'\bisundefined\b', f.read()) is not None
+
+    args = ['murphi2c', '--header', self.model]
+    if self.config['HAS_VALGRIND']:
+      args = ['valgrind', '--leak-check=full', '--show-leak-kinds=all',
+        '--error-exitcode=42'] + args
+    ret, stdout, stderr = run(args, self.env())
+    if self.config['HAS_VALGRIND']:
+      if ret == 42:
+        return Fail(f'Memory leak:\n{stdout}{stderr}')
+
+    # if rumur was expected to reject this model, we allow murphi2c to fail
+    if self.rumur_exit_code == 0 and not should_fail and ret != 0:
+      return Fail(f'Unexpected murphi2c exit status {ret}:\n{stdout}{stderr}')
+
+    if should_fail and ret == 0:
+      return Fail(f'Unexpected murphi2c exit status {ret}:\n{stdout}{stderr}')
+
+    if ret != 0:
+      return None
+
+    with TemporaryDirectory() as tmp:
+
+      # write the header to a temporary file
+      header = os.path.join(tmp, 'header.h')
+      with open(header, 'wt', encoding='utf-8') as f:
+        f.write(stdout)
+
+      # ask the C compiler if the header is valid
+      main_c = f'#include "{header}"\nint main(void) {{ return 0; }}\n'
+      args = [self.config['CC']] + self.config['C_FLAGS'] + ['-o', os.devnull,
+        '-']
+      ret, stdout, stderr = run(args, self.env(), main_c)
+      if ret != 0:
+        return Fail(f'C compilation failed:\n{stdout}{stderr}')
+
+      # ask the C++ compiler if it is valid there too
+      ret, stdout, stderr = run([self.config['CXX'], '-std=c++11', '-o',
+        os.devnull, '-x', 'c++', '-', '-Werror=format', '-Werror=sign-compare',
+        '-Werror=type-limits'], self.env(), main_c)
+      if ret != 0:
+        return Fail(f'C++ compilation failed:\n{stdout}{stderr}')
+
+class Murphi2XMLTest(Tweakable):
+  def __init__(self, model: str, config: Dict[str, Any]):
+    super().__init__(config)
+    self.model = model
+    self.xml = False # dummy setting that tests might reference
+  def description(self) -> str:
+    return f'----{"V" if self.config["HAS_VALGRIND"] else " "} ' \
            f'murphi2xml {os.path.basename(self.model)}'
   def run(self) -> Result:
 
     self.apply_options(self.model)
 
     args = ['murphi2xml', self.model]
-    if HAS_VALGRIND:
+    if self.config['HAS_VALGRIND']:
       args = ['valgrind', '--leak-check=full', '--show-leak-kinds=all',
         '--error-exitcode=42'] + args
-    ret, stdout, stderr = run(args)
-    if HAS_VALGRIND:
+    ret, stdout, stderr = run(args, self.env())
+    if self.config['HAS_VALGRIND']:
       if ret == 42:
         return Fail(f'Memory leak:\n{stdout}{stderr}')
 
@@ -343,12 +353,12 @@ class Murphi2XMLTest(Tweakable):
     xmlcontent = stdout
 
     # See if we have xmllint
-    if which('xmllint') is None:
+    if not self.config['HAS_XMLLINT']:
       return Skip('xmllint not available for validation')
 
     # Validate the XML
     ret, stdout, stderr = run(['xmllint', '--relaxng', MURPHI2XML_RNG,
-      '--noout', '-'], xmlcontent)
+      '--noout', '-'], self.env(), xmlcontent)
     if ret != 0:
       return Fail(f'Failed to validate:\n{stdout}{stderr}')
 
@@ -380,6 +390,19 @@ def main(args: [str]) -> int:
     help='specific test case(s) to run')
   options = parser.parse_args(args[1:])
 
+  # parse configuration
+  config: Dict[str, Any] = {}
+  for p in sorted(os.listdir(os.path.join(os.path.dirname(__file__), 'config'))):
+    path = os.path.join(os.path.dirname(__file__), 'config', p)
+
+    # skip subdirectories
+    if os.path.isdir(path): continue
+
+    # skip non-executable files
+    if not os.access(path, os.X_OK): continue
+
+    config[p] = eval(sp.check_output([path]))
+
   pool = multiprocessing.Pool(options.jobs)
 
   index = 1
@@ -402,7 +425,7 @@ def main(args: [str]) -> int:
     # if this is executable, treat it as a test case
     if os.access(path, os.X_OK):
       if in_range(index):
-        tests.append(Executable(path))
+        tests.append(Executable(path, config))
       index += 1
 
     # if this is not a model, skip the remaining generic logic
@@ -415,17 +438,34 @@ def main(args: [str]) -> int:
       if debug and xml: continue
 
       if in_range(index):
-        tests.append(Model(path, debug, optimised, multithreaded, xml))
+        tests.append(Model(path, debug, optimised, multithreaded, xml, config))
       index += 1
 
     if in_range(index):
-      tests.append(Murphi2XMLTest(path))
+      tests.append(Murphi2XMLTest(path, config))
+    index += 1
+
+    if in_range(index):
+      tests.append(Murphi2CTest(path, config))
+    index += 1
+
+    if in_range(index):
+      tests.append(Murphi2CHeaderTest(path, config))
     index += 1
 
   if len(tests) == 0:
     pr(f'no tests found\n')
     return -1
 
+  pr( 'Configuration:\n'
+     f'  CPUS = {CPUS}\n'
+     f'  STDOUT_ISATTY = {STDOUT_ISATTY}\n'
+     f'  MIN_TEST = {MIN_TEST}\n'
+     f'  MAX_TEST = {MAX_TEST}\n')
+  for key, value in config.items():
+    pr(f'  {key} = {value}\n')
+  pr('\n')
+
   pr(f'Running {len(tests)} tests using {options.jobs} threads...\n'
       '     +------ debug\n'
       '     |+----- optimised\n'
